@@ -3,7 +3,7 @@
 //!
 //! ## Usage
 //!
-//! By default, this iterator is implemented for `[T]`, meaning it works for both arrays and Vec of any type. Chunks must be an owned `Vec<usize>`.
+//! By default, this iterator is implemented for `[T]`, meaning it works for both arrays and Vec of any type.
 //!
 //! If there is not enough data to satisfy the provided chunk length, you will get all the remaining data for that chunk, so it will be shorter than expected. To instead stop early if there is not enough data, see the `.arbitrary_chunks_exact()` variant.
 //!
@@ -14,7 +14,7 @@
 //! let data: Vec<i32> = vec![0, 1, 2, 3, 4];
 //!
 //! let chunked_data: Vec<Vec<i32>> = data
-//!   .arbitrary_chunks(chunks)
+//!   .arbitrary_chunks(&chunks)
 //!   .map(|chunk| chunk.to_vec())
 //!   .collect();
 //!
@@ -32,7 +32,7 @@
 //!
 //! let chunks: Vec<usize> = vec![1, 3];
 //! let data: Vec<i32> = vec![0, 1, 2];
-//! let mut iter = data.arbitrary_chunks_exact(chunks);
+//! let mut iter = data.arbitrary_chunks_exact(&chunks);
 //!
 //! assert_eq!(vec![0], iter.next().unwrap());
 //! assert_eq!(None, iter.next());
@@ -65,7 +65,7 @@
 //! let data: Vec<i32> = vec![0, 1, 2, 3, 4];
 //!
 //! data
-//!   .arbitrary_chunks(chunks)
+//!   .arbitrary_chunks(&chunks)
 //!   .par_bridge()
 //!   .for_each(|chunk| {
 //!     assert!(chunk.len() >= 1 && chunk.len() <= 3);
@@ -99,117 +99,153 @@
 use std::cmp::min;
 use std::mem;
 
-pub struct ArbitraryChunkMut<'a, T: 'a>(&'a mut [T], Vec<usize>);
+pub struct ArbitraryChunkMut<'a, 'b, T: 'a> {
+    data: &'a mut [T],
+    counts: &'b [usize],
+    cursor: usize,
+}
 
-impl<'a, T> Iterator for ArbitraryChunkMut<'a, T> {
+impl<'a, 'b, T> Iterator for ArbitraryChunkMut<'a, 'b, T> {
     type Item = &'a mut [T];
 
     fn next(&mut self) -> Option<Self::Item> {
-        let c = self.1.pop()?;
+        if self.cursor > self.counts.len() - 1 {
+            return None;
+        }
+
+        let c = self.counts[self.cursor];
+        self.cursor += 1;
 
         if c == 0 {
             return Some(&mut []);
         }
 
-        if self.0.is_empty() {
+        if self.data.is_empty() {
             return None;
         }
 
-        let point = min(c, self.0.len());
-        let slice = mem::replace(&mut self.0, &mut []);
+        let point = min(c, self.data.len());
+        let slice = mem::take(&mut self.data);
         let (l, r) = slice.split_at_mut(point);
-        self.0 = r;
+        self.data = r;
 
         Some(l)
     }
 }
 
-pub struct ArbitraryChunk<'a, T: 'a>(&'a [T], Vec<usize>);
+pub struct ArbitraryChunk<'a, 'b, T: 'a> {
+    data: &'a [T],
+    counts: &'b [usize],
+    cursor: usize,
+}
 
-impl<'a, T> Iterator for ArbitraryChunk<'a, T> {
+impl<'a, 'b, T> Iterator for ArbitraryChunk<'a, 'b, T> {
     type Item = &'a [T];
 
     fn next(&mut self) -> Option<Self::Item> {
-        let c = self.1.pop()?;
+        if self.cursor > self.counts.len() - 1 {
+            return None;
+        }
+
+        let c = self.counts[self.cursor];
+        self.cursor += 1;
 
         if c == 0 {
             return Some(&[]);
         }
 
-        if self.0.is_empty() {
+        if self.data.is_empty() {
             return None;
         }
 
-        let point = min(c, self.0.len());
-        let slice = mem::replace(&mut self.0, &mut []);
+        let point = min(c, self.data.len());
+        let slice = mem::take(&mut self.data);
         let (l, r) = slice.split_at(point);
-        self.0 = r;
+        self.data = r;
 
         Some(l)
     }
 }
 
-pub struct ArbitraryChunkExactMut<'a, T: 'a>(&'a mut [T], Vec<usize>);
+pub struct ArbitraryChunkExactMut<'a, 'b, T: 'a> {
+    data: &'a mut [T],
+    counts: &'b [usize],
+    cursor: usize,
+}
 
-impl<'a, T> Iterator for ArbitraryChunkExactMut<'a, T> {
+impl<'a, 'b, T> Iterator for ArbitraryChunkExactMut<'a, 'b, T> {
     type Item = &'a mut [T];
 
     fn next(&mut self) -> Option<Self::Item> {
-        let c = self.1.pop()?;
+        if self.cursor > self.counts.len() - 1 {
+            return None;
+        }
+
+        let c = self.counts[self.cursor];
+        self.cursor += 1;
 
         if c == 0 {
             return Some(&mut []);
         }
 
-        if self.0.is_empty() || c > self.0.len() {
+        if self.data.is_empty() || c > self.data.len() {
             return None;
         }
 
-        let slice = mem::replace(&mut self.0, &mut []);
+        let slice = mem::take(&mut self.data);
         let (l, r) = slice.split_at_mut(c);
-        self.0 = r;
+        self.data = r;
 
         Some(l)
     }
 }
 
-impl<'a, T> ArbitraryChunkExactMut<'a, T> {
+impl<'a, 'b, T> ArbitraryChunkExactMut<'a, 'b, T> {
     pub fn remainder(&'a mut self) -> &'a mut [T] {
-        self.0
+        self.data
     }
 }
 
-pub struct ArbitraryChunkExact<'a, T: 'a>(&'a [T], Vec<usize>);
+pub struct ArbitraryChunkExact<'a, 'b, T: 'a> {
+    data: &'a [T],
+    counts: &'b [usize],
+    cursor: usize,
+}
 
-impl<'a, T> Iterator for ArbitraryChunkExact<'a, T> {
+impl<'a, 'b, T> Iterator for ArbitraryChunkExact<'a, 'b, T> {
     type Item = &'a [T];
 
     fn next(&mut self) -> Option<Self::Item> {
-        let c = self.1.pop()?;
+        if self.cursor > self.counts.len() - 1 {
+            return None;
+        }
+
+        let c = self.counts[self.cursor];
+        self.cursor += 1;
 
         if c == 0 {
             return Some(&[]);
         }
 
-        if self.0.is_empty() || c > self.0.len() {
+        if self.data.is_empty() || c > self.data.len() {
             return None;
         }
 
-        let slice = mem::replace(&mut self.0, &mut []);
+        let slice = mem::take(&mut self.data);
         let (l, r) = slice.split_at(c);
-        self.0 = r;
+        self.data = r;
 
         Some(l)
     }
 }
 
-impl<'a, T> ArbitraryChunkExact<'a, T> {
+impl<'a, 'b, T> ArbitraryChunkExact<'a, 'b, T> {
     pub fn remainder(&'a self) -> &'a [T] {
-        self.0
+        self.data
     }
 }
 
-pub trait ArbitraryChunks<T> {
+pub trait ArbitraryChunks<'a, 'b, T> {
     /// `arbitrary_chunks` returns an iterator over chunks of sizes defined in `counts`.
     ///
     /// ```rust
@@ -219,7 +255,7 @@ pub trait ArbitraryChunks<T> {
     /// let data: Vec<i32> = vec![0, 1, 2, 3, 4];
     ///
     /// let chunked_data: Vec<Vec<i32>> = data
-    ///     .arbitrary_chunks(chunks)
+    ///     .arbitrary_chunks(&chunks)
     ///     .map(|chunk| chunk.to_vec())
     ///     .collect();
     ///
@@ -227,7 +263,7 @@ pub trait ArbitraryChunks<T> {
     /// assert_eq!(vec![1, 2, 3], chunked_data[1]);
     /// assert_eq!(vec![4], chunked_data[2]);
     /// ```
-    fn arbitrary_chunks(&self, counts: Vec<usize>) -> ArbitraryChunk<T>;
+    fn arbitrary_chunks(&'a self, counts: &'b [usize]) -> ArbitraryChunk<'a, 'b, T>;
 
     /// `arbitrary_chunks_mut` returns an iterator over mutable chunks of sizes defined in `counts`.
     ///
@@ -238,14 +274,14 @@ pub trait ArbitraryChunks<T> {
     /// let mut data: Vec<i32> = vec![0, 1, 2, 3, 4];
     ///
     /// data
-    ///     .arbitrary_chunks_mut(chunks)
+    ///     .arbitrary_chunks_mut(&chunks)
     ///     .for_each(|chunk| {
     ///         chunk[0] = chunk[0] * 2;
     ///     });
     ///
     /// assert_eq!(vec![0, 2, 2, 3, 8], data);
     /// ```
-    fn arbitrary_chunks_mut(&mut self, counts: Vec<usize>) -> ArbitraryChunkMut<T>;
+    fn arbitrary_chunks_mut(&'a mut self, counts: &'b [usize]) -> ArbitraryChunkMut<'a, 'b, T>;
 
     /// `arbitrary_chunks_exact` returns chunks sized exactly as requested, or not at all.
     /// If there is not enough data to satisfy the chunk, the iterator will end. You will then be
@@ -256,13 +292,13 @@ pub trait ArbitraryChunks<T> {
     ///
     /// let chunks: Vec<usize> = vec![1, 3];
     /// let data: Vec<i32> = vec![0, 1, 2];
-    /// let mut iter = data.arbitrary_chunks_exact(chunks);
+    /// let mut iter = data.arbitrary_chunks_exact(&chunks);
     ///
     /// assert_eq!(vec![0], iter.next().unwrap());
     /// assert_eq!(None, iter.next());
     /// assert_eq!(vec![1, 2], iter.remainder());
     /// ```
-    fn arbitrary_chunks_exact(&self, counts: Vec<usize>) -> ArbitraryChunkExact<T>;
+    fn arbitrary_chunks_exact(&'a self, counts: &'b [usize]) -> ArbitraryChunkExact<'a, 'b, T>;
 
     /// `arbitrary_chunks_exact_mut` returns chunks sized exactly as requested, or not at all.
     /// If there is not enough data to satisfy the chunk, the iterator will end. You will then be
@@ -273,34 +309,46 @@ pub trait ArbitraryChunks<T> {
     ///
     /// let chunks: Vec<usize> = vec![1, 3];
     /// let mut data: Vec<i32> = vec![0, 1, 2];
-    /// let mut iter = data.arbitrary_chunks_exact_mut(chunks);
+    /// let mut iter = data.arbitrary_chunks_exact_mut(&chunks);
     ///
     /// assert_eq!(vec![0], iter.next().unwrap());
     /// assert_eq!(None, iter.next());
     /// assert_eq!(vec![1, 2], iter.remainder());
     /// ```
-    fn arbitrary_chunks_exact_mut(&mut self, counts: Vec<usize>) -> ArbitraryChunkExactMut<T>;
+    fn arbitrary_chunks_exact_mut(&'a mut self, counts: &'b [usize]) -> ArbitraryChunkExactMut<'a, 'b, T>;
 }
 
-impl<T> ArbitraryChunks<T> for [T] {
-    fn arbitrary_chunks(&self, mut counts: Vec<usize>) -> ArbitraryChunk<T> {
-        counts.reverse();
-        ArbitraryChunk(self, counts)
+impl<'a, 'b, T> ArbitraryChunks<'a, 'b, T> for [T] {
+    fn arbitrary_chunks(&'a self, counts: &'b [usize]) -> ArbitraryChunk<'a, 'b, T> {
+        ArbitraryChunk {
+            data: self,
+            counts,
+            cursor: 0,
+        }
     }
 
-    fn arbitrary_chunks_mut(&mut self, mut counts: Vec<usize>) -> ArbitraryChunkMut<T> {
-        counts.reverse();
-        ArbitraryChunkMut(self, counts)
+    fn arbitrary_chunks_mut(&'a mut self, counts: &'b [usize]) -> ArbitraryChunkMut<'a, 'b, T> {
+        ArbitraryChunkMut {
+            data: self,
+            counts,
+            cursor: 0,
+        }
     }
 
-    fn arbitrary_chunks_exact(&self, mut counts: Vec<usize>) -> ArbitraryChunkExact<T> {
-        counts.reverse();
-        ArbitraryChunkExact(self, counts)
+    fn arbitrary_chunks_exact(&'a self, counts: &'b [usize]) -> ArbitraryChunkExact<'a, 'b, T> {
+        ArbitraryChunkExact {
+            data: self,
+            counts,
+            cursor: 0,
+        }
     }
 
-    fn arbitrary_chunks_exact_mut(&mut self, mut counts: Vec<usize>) -> ArbitraryChunkExactMut<T> {
-        counts.reverse();
-        ArbitraryChunkExactMut(self, counts)
+    fn arbitrary_chunks_exact_mut(&'a mut self, counts: &'b [usize]) -> ArbitraryChunkExactMut<'a, 'b, T> {
+        ArbitraryChunkExactMut {
+            data: self,
+            counts,
+            cursor: 0,
+        }
     }
 }
 
@@ -313,7 +361,7 @@ mod tests {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let data = vec![8, 7, 6, 5, 4, 3, 2, 1];
         let chunk_data: Vec<Vec<i32>> = data
-            .arbitrary_chunks(chunks)
+            .arbitrary_chunks(&chunks)
             .map(|chunk| chunk.to_vec())
             .collect();
 
@@ -329,7 +377,7 @@ mod tests {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let mut data = vec![8, 7, 6, 5, 4, 3, 2, 1];
         let chunk_data: Vec<Vec<i32>> = data
-            .arbitrary_chunks_mut(chunks)
+            .arbitrary_chunks_mut(&chunks)
             .map(|chunk| chunk.to_vec())
             .collect();
 
@@ -344,7 +392,7 @@ mod tests {
     fn exact_stops_when_chunks_run_out() {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let data = vec![8, 7, 6, 5, 4, 3, 2, 1];
-        let mut iter = data.arbitrary_chunks_exact(chunks);
+        let mut iter = data.arbitrary_chunks_exact(&chunks);
 
         assert_eq!(&[0i32; 0], iter.next().unwrap());
         assert_eq!(&[8i32], iter.next().unwrap());
@@ -358,7 +406,7 @@ mod tests {
     fn exact_mut_stops_when_chunks_run_out() {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let mut data = vec![8, 7, 6, 5, 4, 3, 2, 1];
-        let mut iter = data.arbitrary_chunks_exact_mut(chunks);
+        let mut iter = data.arbitrary_chunks_exact_mut(&chunks);
 
         assert_eq!(&mut [0i32; 0], iter.next().unwrap());
         assert_eq!(&mut [8i32], iter.next().unwrap());
@@ -373,7 +421,7 @@ mod tests {
         let chunks: Vec<usize> = vec![0, 1, 2, 3, 0, 0];
         let data = vec![8, 7, 6, 5, 4, 3, 2, 1];
         let chunk_data: Vec<Vec<i32>> = data
-            .arbitrary_chunks(chunks)
+            .arbitrary_chunks(&chunks)
             .map(|chunk| chunk.to_vec())
             .collect();
 
@@ -391,7 +439,7 @@ mod tests {
         let chunks: Vec<usize> = vec![0, 1, 2, 3, 0, 0];
         let mut data = vec![8, 7, 6, 5, 4, 3, 2, 1];
         let chunk_data: Vec<Vec<i32>> = data
-            .arbitrary_chunks_mut(chunks)
+            .arbitrary_chunks_mut(&chunks)
             .map(|chunk| chunk.to_vec())
             .collect();
 
@@ -408,7 +456,7 @@ mod tests {
     fn exact_accounts_for_trailing_zeros() {
         let chunks: Vec<usize> = vec![0, 1, 2, 3, 0, 0];
         let data = vec![8, 7, 6, 5, 4, 3, 2, 1];
-        let mut iter = data.arbitrary_chunks_exact(chunks);
+        let mut iter = data.arbitrary_chunks_exact(&chunks);
 
         assert_eq!(&[0i32; 0], iter.next().unwrap());
         assert_eq!(&[8i32], iter.next().unwrap());
@@ -424,7 +472,7 @@ mod tests {
     fn exact_mut_accounts_for_trailing_zeros() {
         let chunks: Vec<usize> = vec![0, 1, 2, 3, 0, 0];
         let mut data = vec![8, 7, 6, 5, 4, 3, 2, 1];
-        let mut iter = data.arbitrary_chunks_exact_mut(chunks);
+        let mut iter = data.arbitrary_chunks_exact_mut(&chunks);
 
         assert_eq!(&mut [0i32; 0], iter.next().unwrap());
         assert_eq!(&mut [8i32], iter.next().unwrap());
@@ -442,7 +490,7 @@ mod tests {
         let data = vec![8, 7, 6, 5, 4];
 
         let chunk_data: Vec<Vec<i32>> = data
-            .arbitrary_chunks(chunks)
+            .arbitrary_chunks(&chunks)
             .map(|chunk| chunk.to_vec())
             .collect();
 
@@ -458,7 +506,7 @@ mod tests {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let mut data = vec![8, 7, 6, 5, 4];
         let chunk_data: Vec<Vec<i32>> = data
-            .arbitrary_chunks_mut(chunks)
+            .arbitrary_chunks_mut(&chunks)
             .map(|chunk| chunk.to_vec())
             .collect();
 
@@ -473,7 +521,7 @@ mod tests {
     fn exact_stops_when_data_runs_out() {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let data = vec![8, 7, 6, 5, 4];
-        let mut iter = data.arbitrary_chunks_exact(chunks);
+        let mut iter = data.arbitrary_chunks_exact(&chunks);
 
         assert_eq!(&[0i32; 0], iter.next().unwrap());
         assert_eq!(&[8i32], iter.next().unwrap());
@@ -486,7 +534,7 @@ mod tests {
     fn exact_mut_stops_when_data_runs_out() {
         let chunks: Vec<usize> = vec![0, 1, 2, 3];
         let mut data = vec![8, 7, 6, 5, 4];
-        let mut iter = data.arbitrary_chunks_exact_mut(chunks);
+        let mut iter = data.arbitrary_chunks_exact_mut(&chunks);
 
         assert_eq!(&mut [0i32; 0], iter.next().unwrap());
         assert_eq!(&mut [8i32], iter.next().unwrap());
